;;;; support.scm - Support code for structures
;
; Copyright (c) 2000-2003, Felix L. Winkelmann
; All rights reserved.
;
; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
; conditions are met:
;
;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
;     disclaimer. 
;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
;     disclaimer in the documentation and/or other materials provided with the distribution. 
;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
;     products derived from this software without specific prior written permission. 
;
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
; POSSIBILITY OF SUCH DAMAGE.
;
; Send bugs, suggestions and ideas to: 
;
; felix@call-with-current-continuation.org
;
; Felix L. Winkelmann
; Steinweg 1A
; 37130 Gleichen, OT Weissenborn
; Germany

(declare
  (export structures:signature-table
	  structures:make-structure structures:restrict-structure
	  structures:structure-ref structures:structure?
	  open structure-ref compound-structure structure?)
  (uses srfi-1 extras)				; eval?
  (fixnum) )

(define structures:signature-table (make-hash-table eq?))

(define (open . structs)
  (for-each
   (lambda (s)
     (##sys#check-structure s 'structure 'open)
     (for-each (lambda (b) (##sys#setslot (car b) 0 (cdr b))) (##sys#slot s 1)) )
   structs) )

(define (structures:structure-ref struct name)
  (##sys#check-structure struct 'structure 'structure-ref)
  (##sys#check-symbol name 'structure-ref)
  (or (and-let* ([a (assq name (##sys#slot struct 1))]) (cdr a))
      (error "structure does not export binding" struct name) ) )

(define structure-ref structures:structure-ref)

(define (compound-structure . structs)
  (define (binding-eq? b1 b2) (eq? (car b1) (car b2)))
  (##sys#make-structure
   'structure
   (let loop ([structs structs] [bs '()])
     (if (null? structs)
	 bs
	 (let ([s (car structs)])
	   (##sys#check-structure s 'structure 'compound-structure)
	   (let* ([bs2 (##sys#slot s 1)]
		  [in (lset-intersection binding-eq? bs bs2)] )
	     (if (null? in)
		 (loop (cdr structs) (append bs2 bs))
		 (error "bindings multiply defined in compound structure" (unzip1 in)) ) ) ) ) ) ) )

(define (structures:structure? x . exports)
  (and (##sys#structure? x 'structure)
       (let ([bs (##sys#slot x 1)])
	 (every 
	  (match-lambda 
	    [('export exp ...)
	     (every
	      (match-lambda 
		[(? symbol? s) (assq s bs)] 
		[x (error "invalid export syntax" x)] ) 
	      exp) ]
	    [(? symbol? sig)
	     (let ([ss (hash-table-ref/default structures:signature-table sig #f)])
	       (if ss
		   (every
		    (match-lambda 
		      [(sym . pred) 
		       (and-let* ([a (assq sym bs)]) (pred (cdr a))) ]
		      [sym (assq sym bs)] )
		    ss)
		   (error "undefined signature" sig) ) ) ]
	    [x (error "invalid export syntax" x)] )
	  exports) ) ) )

(define structure? structures:structure?)

(define (structures:make-structure bindings exports)
  (structures:restrict-structure (##sys#make-structure 'structure bindings) exports) )

(define (structures:restrict-structure s exports)
  (unless (null? exports)
    (##sys#setslot s 1 (restrict-bindings (##sys#slot s 1) exports)) )
  s)

(define (restrict-bindings bs exports)
  (let ([found '()])
    (define (test b)
      (let ([a (assq b bs)])
	(if a
	    (set! found (cons a found)) 
	    (error "structure does not export required binding" b) ) ) )
    (define (test* b pred)
      (let ([a (assq b bs)])
	(if (and a (pred (cdr a)))
	    (set! found (cons a found))
	    (error "structure does not export required binding or type does not match" b) ) ) )
    (for-each
     (match-lambda
       [('export exp ...)
	(for-each 
	 (match-lambda
	   [(? symbol? s) (test s)] 
	   [x (error "invalid export syntax" x)] )
	 exp) ]
       [(? symbol? sig)
	(let ([ss (hash-table-ref/default structures:signature-table sig #f)])
	  (if ss
	      (for-each
	       (match-lambda 
		 [(sym . pred) (test* sym pred)]
		 [sym (test sym)] )
	       ss)
	      (error "undefined signature" sig) ) ) ]
       [x (error "invalid export syntax" x)] )
     exports)
    found) )

